# -*- coding: utf-8 -*-

"""

tests.test_commons

Unit test the postleid.commons module

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


# import io
# import itertools
import logging

# import os
import random

# import secrets
# import tempfile
# import time

from unittest import TestCase

# from unittest.mock import patch

from postleid import commons

# from .commons import GenericCallResult


# OUT_OF_RANGE = postleid.formal_postal_code_checks.OutOfRangeError
# INVALID_FORMAT = postleid.formal_postal_code_checks.InvalidFormatError


# =============================================================================
# class ExecResult(GenericCallResult):
#
#     """Program execution result"""
#
#     @classmethod
#     def do_call(cls, *args, **kwargs):
#         """Do the real function call"""
#         return postleid.main(list(args))
#
# =============================================================================


class Fuctions(TestCase):

    """Helper functions"""

    def test_log_formatted(self):
        """LogWrapper.log_formatted function"""
        for loglevel in (
            logging.DEBUG,
            logging.INFO,
            logging.WARNING,
            logging.ERROR,
        ):
            level_name = logging.getLevelName(loglevel)
            with self.subTest(level=level_name):
                nb_messages = random.randint(1, 5)
                expected_messages = []
                with self.assertLogs(level=level_name) as log_cm:
                    for message_no in range(nb_messages):
                        expected_messages.append(
                            f"{level_name}:root:"
                            f"testing {level_name} #{message_no}"
                        )
                        commons.LogWrapper.log_formatted(
                            loglevel, f"testing {level_name} #{message_no}"
                        )
                    #
                #
                for message_no in range(nb_messages):
                    self.assertEqual(log_cm.output, expected_messages)
                #
            #
        #

    def test_separator_line(self):
        """separator_line function"""
        for width in range(16, 32, 77):
            for codepoint in range(1024):
                element = chr(codepoint)
                with self.subTest(element=element, width=width):
                    self.assertEqual(
                        commons.separator_line(element=element, width=width),
                        element * width,
                    )
                #
            #
        #


# TODO: - create a test document from scratch and use that to test
#         all possible test cases → in a new test module,
#         eg. test_postleid_main
#       - open the output document and ensure all possible conversions
#         did really happen

# =============================================================================
# class Program(TestCase):
#
#     """Test the Program class"""
#
#     def test_complete_format(self):
#         """Test completing a format, all combinations"""
#         for target_format in ("input", "toggle", "JSON", "YAML"):
#             combinations = [
#                 (char.lower(), char.upper()) for char in target_format
#             ]
#             matching_abbreviations = set()
#             for index in range(len(target_format)):
#                 matching_abbreviations.update(
#                     set(itertools.product(*combinations[: index + 1]))
#                 )
#             #
#             for current_input in matching_abbreviations:
#                 format_str = "".join(current_input)
#                 with self.subTest(format_str=format_str):
#                     self.assertEqual(
#                         commandline.Program.complete_format(format_str),
#                         target_format,
#                     )
#                 #
#             #
#         #
#
#     def test_unsupported_format(self):
#         """Test unsupported formats"""
#         for target_format in ("XML", "xml", "ini", "toml"):
#             with self.subTest(target_format=target_format):
#                 self.assertRaisesRegex(
#                     ValueError,
#                     f"Unsupported format {target_format!r}!",
#                     commandline.Program.complete_format,
#                     target_format,
#                 )
#             #
#         #
#         #
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_stdin_extract(self, mock_stdout):
#         """Test execute function, reading stdin, extract mode"""
#         with self.assertLogs(level="INFO") as log_cm:
#             result = ExecResult.from_call(
#                 "--verbose",
#                 stdin_data=data.INPUT_YAML_ANIMALS,
#                 stdout=mock_stdout,
#                 data_path=access.ExtractingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent("big"),
#                     access.PathComponent(1, in_subscript=True),
#                 ),
#             )
#         self.assertEqual(
#             log_cm.records[0].message, "Operating in extract mode."
#         )
#         self.assertEqual(result.stdout, data.EXPECT_CATS_BIG_1)
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_stdin_replace(self, mock_stdout):
#         """Test execute function, reading stdin, replace mode"""
#         with self.assertLogs(level="INFO") as log_cm:
#             result = ExecResult.from_call(
#                 "--debug",
#                 stdin_data=data.INPUT_YAML_ANIMALS,
#                 stdout=mock_stdout,
#                 data_path=access.ReplacingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent("big"),
#                     access.PathComponent(2, in_subscript=True),
#                     replacement="panther",
#                 ),
#             )
#         self.assertEqual(
#             log_cm.records[0].message, "Operating in replace mode."
#         )
#         self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_stdin_passthru(self, mock_stdout):
#         """Test execute function, reading stdin, extract mode (passthru)"""
#         with self.assertLogs(level="INFO") as log_cm:
#             result = ExecResult.from_call(
#                 "--verbose",
#                 stdin_data=data.INPUT_YAML_ANIMALS,
#                 stdout=mock_stdout,
#             )
#         #
#         self.assertEqual(
#             log_cm.records[0].message,
#             "Operating in extract mode (passthrough).",
#         )
#         self.assertEqual(result.stdout, data.INPUT_YAML_ANIMALS)
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_invalid_input(self, mock_stdout):
#         """Test execute function, reading stdin, invalid input"""
#         with self.assertLogs(level="ERROR") as log_cm:
#             result = ExecResult.from_call(
#                 stdin_data="a: b: c",
#                 stdout=mock_stdout,
#             )
#         #
#         self.assertIn(
#             # FIXME: PyYAML specific error message, might change
#             "mapping values are not allowed here",
#             log_cm.records[0].message,
#         )
#         self.assertEqual(result.stdout, "")
#         self.assertEqual(result.returncode, data.RETURNCODE_ERROR)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_not_found(self, mock_stdout):
#         """Test execute function, reading stdin, value not found"""
#         wrong_index = "HUGE"
#         with self.assertLogs(level="ERROR") as log_cm:
#             result = ExecResult.from_call(
#                 stdin_data=data.INPUT_YAML_ANIMALS,
#                 stdout=mock_stdout,
#                 data_path=access.ReplacingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent(wrong_index),
#                 ),
#             )
#         #
#         self.assertEqual(
#             log_cm.records[0].message,
#             f"Index {wrong_index!r} not found!",
#         )
#         self.assertEqual(result.stdout, "")
#         self.assertEqual(result.returncode, data.RETURNCODE_ERROR)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_stdin_inplace(self, mock_stdout):
#         """Test execute function, reading stdin,
#         replace mode with --inplace"""
#         with self.assertLogs(level="WARNING") as log_cm:
#             result = ExecResult.from_call(
#                 "--inplace",
#                 stdin_data=data.INPUT_YAML_ANIMALS,
#                 stdout=mock_stdout,
#                 data_path=access.ReplacingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent("big"),
#                     access.PathComponent(2, in_subscript=True),
#                     replacement="panther",
#                 ),
#             )
#         self.assertEqual(
#             log_cm.records[0].message, "Cannot modify <stdin> in place"
#         )
#         self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_file_read(self, mock_stdout):
#         """Test reading a file"""
#         with tempfile.TemporaryDirectory() as temporary_directory:
#             yaml_file_name = os.path.join(
#                 temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
#             )
#             with open(
#                 yaml_file_name, mode="w", encoding="utf-8"
#             ) as source_file:
#                 source_file.write(data.INPUT_YAML_ANIMALS)
#             #
#             time.sleep(0.1)
#             creation_time = os.stat(yaml_file_name).st_mtime
#             result = ExecResult.from_call(
#                 "",
#                 yaml_file_name,
#                 stdout=mock_stdout,
#                 data_path=access.ReplacingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent("big"),
#                     access.PathComponent(2, in_subscript=True),
#                     replacement="panther",
#                 ),
#             )
#             check_time = os.stat(yaml_file_name).st_mtime
#         #
#         self.assertEqual(result.stdout, data.EXPECT_YAML_JAGUAR2PANTHER)
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#         self.assertEqual(check_time, creation_time)
#
#     @patch("sys.stdout", new_callable=io.StringIO)
#     def test_exec_file_write(self, mock_stdout):
#         """Test modifying a file in place"""
#         with tempfile.TemporaryDirectory() as temporary_directory:
#             yaml_file_name = os.path.join(
#                 temporary_directory, f"{secrets.token_urlsafe(6)}.yaml"
#             )
#             with open(
#                 yaml_file_name, mode="w", encoding="utf-8"
#             ) as source_file:
#                 source_file.write(data.INPUT_YAML_ANIMALS)
#             #
#             time.sleep(0.1)
#             creation_time = os.stat(yaml_file_name).st_mtime
#             result = ExecResult.from_call(
#                 "--inplace",
#                 "",
#                 yaml_file_name,
#                 stdout=mock_stdout,
#                 data_path=access.ReplacingPath(
#                     access.PathComponent("felines"),
#                     access.PathComponent("cats"),
#                     access.PathComponent("big"),
#                     access.PathComponent(2, in_subscript=True),
#                     replacement="panther",
#                 ),
#             )
#             check_time = os.stat(yaml_file_name).st_mtime
#             self.assertGreater(check_time, creation_time)
#             with open(
#                 yaml_file_name, mode="r", encoding="utf-8"
#             ) as changed_file:
#                 written_data = changed_file.read()
#             #
#             self.assertEqual(written_data, data.EXPECT_YAML_JAGUAR2PANTHER)
#         #
#         self.assertEqual(result.stdout, "")
#         self.assertEqual(result.returncode, data.RETURNCODE_OK)
#
# =============================================================================

# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



# -*- coding: utf-8 -*-

"""

tests.test_rule_checks

Unit test the postleid.rule_checks module

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from unittest import TestCase

from postleid import rule_checks


InvalidFormat = rule_checks.InvalidFormatError
OutOfRange = rule_checks.OutOfRangeError
MissingRules = rule_checks.MissingRulesError

CORRECT = [
    ("ad", "ad123", "AD123"),
    ("af", 1234, "1234"),
    ("af", "3456", "3456"),
    ("ai", "ai-2640", "AI-2640"),
    ("al", "1001", "1001"),
    ("al", 9799, "9799"),
    ("am", 4299, "4299"),
    ("aq", "biqq 1zz", "BIQQ 1ZZ"),
    ("ar", 2345, "2345"),
    ("ar", "A2345BCD", "A2345BCD"),
    ("as", 96799, "96799"),
    ("as", "96799-1234", "96799-1234"),
    ("at", 9999, "9999"),
    ("au", 200, "0200"),
    ("au", 9999, "9999"),
    ("ax", "ax-22000", "AX-22000"),
    ("az", "AZ 0123", "AZ 0123"),
    ("az", "az 6999", "AZ 6999"),
    ("bb", "bb11000", "BB11000"),
    ("bd", "1000", "1000"),
    ("bd", 9499, "9499"),
    ("bh", 101, "101"),
    ("bh", "101", "101"),
    ("bh", 1216, "1216"),
    ("bm", "ab cd", "AB CD"),
    ("bm", "ab 12", "AB 12"),
]

ERRORS = [
    ("xw", 12345, MissingRules),
    ("ad", "AD099", OutOfRange),
    ("ad", "AD800", OutOfRange),
    ("ad", "123", InvalidFormat),
    ("af", 999, OutOfRange),
    ("af", "4400", OutOfRange),
    ("al", 999, OutOfRange),
    ("al", "9800", OutOfRange),
    ("am", 4300, OutOfRange),
    ("am", 12345, InvalidFormat),
    ("ai", "2640", InvalidFormat),
    ("ai", "AI-2222", InvalidFormat),
    ("aq", "BIQQ 2ZZ", InvalidFormat),
    ("ar", 345, InvalidFormat),
    ("ar", "A0345BCD", OutOfRange),
    ("ar", "I2345BCD", InvalidFormat),
    ("as", 96780, InvalidFormat),
    ("as", "96799-ABCD", InvalidFormat),
    ("at", 999, OutOfRange),
    ("au", 199, OutOfRange),
    ("ax", "AX-77000", InvalidFormat),
    ("az", "AZ 0099", OutOfRange),
    ("az", "AZ 7000", OutOfRange),
    ("bb", 11000, InvalidFormat),
    ("bd", "999", InvalidFormat),
    ("bd", 999, OutOfRange),
    ("bh", "100", OutOfRange),
    ("bh", 1217, OutOfRange),
    ("bm", "ABCD", InvalidFormat),
    ("bm", "AB99", InvalidFormat),
    ("bm", 9999, InvalidFormat),
    ("by", "99999", InvalidFormat),
]


class PostalCodes(TestCase):

    """Test postal codes per country"""

    def test_correct(self):
        """correct codes per country"""
        v_cache = rule_checks.ValidatorsCache()
        for country, candidate, expected_result in CORRECT:
            with self.subTest(country=country, candidate=candidate):
                self.assertEqual(
                    v_cache.output_validated(candidate, country=country),
                    expected_result,
                )
            #
        #

    def test_errors(self):
        """wrong codes per country"""
        v_cache = rule_checks.ValidatorsCache()
        for country, candidate, expected_exception in ERRORS:
            with self.subTest(country=country, candidate=candidate):
                self.assertRaises(
                    expected_exception,
                    v_cache.output_validated,
                    candidate,
                    country=country,
                )
            #
        #


# vim: fileencoding=utf-8 ts=4 sts=4 sw=4 autoindent expandtab syntax=python:



#!/usr/bin/env python

"""

postleid.fix_excel_files

Class for fixing postal codes in excel files

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


from typing import Any, Dict, List, Union

import numpy
import pandas

# local imports
from postleid import commons
from postleid import paths
from postleid import rule_checks


class DataFixer:

    """Fix cells in a workbook"""

    cc_column = "country_code"

    results_by_error = {
        rule_checks.InvalidFormatError: commons.S_WRONG_FORMAT,
        rule_checks.MissingRulesError: commons.S_MISSING_RULES,
        rule_checks.OutOfRangeError: commons.S_OUT_OF_RANGE,
        rule_checks.UnsupportedDataTypeError: commons.S_WRONG_DATA_TYPE,
    }

    def __init__(
        self,
        dataframe: pandas.DataFrame,
        user_settings: commons.UserSettings,
    ) -> None:
        """Find the active sheet in the workbook"""
        self.dataframe = dataframe
        self.user_settings = user_settings
        self.zip_column = ""
        self.country_column = ""
        postal_code_heading_parts = [
            part.lower()
            for part in self.user_settings.postal_code_heading_parts
        ]
        country_headings = [
            heading.lower() for heading in self.user_settings.country_headings
        ]
        for column_name in self.dataframe.columns:
            preprocessed_name = column_name.strip().lower()
            if not self.zip_column:
                for name_part in postal_code_heading_parts:
                    if name_part in preprocessed_name:
                        self.zip_column = column_name
                        break
                    #
                #
            #
            if not self.country_column:
                if preprocessed_name in country_headings:
                    self.country_column = column_name
                #
            #
            if column_name == self.cc_column:
                self.cc_column = f"{column_name}_x"
            #
        #
        # Build a cc column
        self.__cc_lookup: Dict[str, str] = {}
        for (
            iso_cc,
            country_names,
        ) in commons.load_country_names_from_file().items():
            for name in country_names:
                self.__cc_lookup[name.lower()] = iso_cc
            #
        #

        if self.country_column:
            commons.LogWrapper.debug(
                f"Reading country data from {self.country_column}"
            )
            country_source = self.dataframe[self.country_column]
        else:
            country_source = pandas.Series(self.dataframe.index)
        #
        # Map ISO-3166-1 ALPHA-2 codes to country names
        # using the self.lookup_country_code() function because
        # a simple dict is not sufficient (that would produce
        # NaN values for unmatched country names), and we need
        # even more flexibility than a collections.defaultdict()
        # could provide.
        self.dataframe[self.cc_column] = country_source.map(
            self.lookup_country_code
        )
        for line in str(self.dataframe).splitlines():
            commons.LogWrapper.debug(line)
        #
        self.__validator = rule_checks.ValidatorsCache(
            default_cc=self.user_settings.default_country_code
        )

    def lookup_country_code(self, country: Union[float, int, str]) -> str:
        """Lookup the country code for the given country"""
        country_code = ""
        if isinstance(country, (float, int)):
            return country_code
        #
        preprocessed_name = country.strip().lower()
        try:
            country_code = self.__cc_lookup[preprocessed_name]
        except KeyError:
            if preprocessed_name:
                country_code = "??"
                commons.LogWrapper.warning(
                    f"Country code für {country!r} nicht gefunden,"
                    f" verwende {country_code!r}."
                )
            #
        #
        if country_code == self.user_settings.default_country_code:
            return ""
        #
        return country_code

    def fix_all_zip_codes(self) -> List[str]:
        """Fix all zip codes in an Excel workbook.
        Returns statistics (a list of keywords).
        """
        statistics: List[str] = []
        for row_number in self.dataframe.index:
            statistics.append(self.fix_single_cell(row_number))
        #
        return statistics

    def fix_single_cell(self, row_number: int) -> str:
        """Fix the zip code in a single cell.
        Delegate fixing to the appropriate operation
        for the cell content type and log the message.
        Return the operation result.
        """
        original_value = self.dataframe.at[row_number, self.zip_column]
        preprocessed_value = self.__preprocess_cell_value(original_value)
        country_code = (
            self.dataframe.at[row_number, self.cc_column]
            or self.user_settings.default_country_code
        )
        error_details: List[str] = []
        result = commons.S_MISSING_RULES
        try:
            new_value = self.__validator.output_validated(
                preprocessed_value, country=country_code
            )
        except rule_checks.ValidatorError as error:
            error_details.extend(error.args)
            error_details.extend(error.additional_information)
            result = self.results_by_error[type(error)]
        #
        if error_details:
            commons.LogWrapper.warning(
                f"{row_number:>5}  →  Originalwert: {original_value!r}",
                f"      {result} - {error_details[0]}",
            )
            commons.LogWrapper.debug(
                *[f"          - {detail}" for detail in error_details[1:]],
                "      --- Typ nach Vorbehandlung:"
                f" {type(preprocessed_value)}",
            )
            return result
        #
        if new_value == original_value:
            result, details = commons.S_UNCHANGED, "keine Anpassung nötig"
        else:
            self.dataframe.at[row_number, self.zip_column] = new_value
            result, details = commons.S_FIXED, f"neuer Wert: {new_value!r}"
        #
        commons.LogWrapper.debug(
            f"{row_number:>5}  →  Originalwert: {original_value!r}",
            f"      {result} – {details}",
        )
        return result

    def __preprocess_cell_value(
        self, original_value: Any
    ) -> Union[float, int, str]:
        """Return a preprocessed variant of the original value"""
        preprocessed_value = original_value
        if isinstance(original_value, str):
            try:
                preprocessed_value = float(original_value.replace(",", "."))
            except ValueError:
                pass
            else:
                commons.LogWrapper.debug(
                    f"       --- {original_value}"
                    f" → Float: {preprocessed_value}"
                )
            #
        #
        if isinstance(preprocessed_value, numpy.integer):
            preprocessed_value = int(preprocessed_value)
        elif isinstance(preprocessed_value, numpy.inexact):
            preprocessed_value = float(preprocessed_value)
        #
        if (
            isinstance(preprocessed_value, (int, float))
            and self.user_settings.guess_1000s
            and preprocessed_value < 1000
        ):
            return preprocessed_value * 1000
        #
        return preprocessed_value

    def sort_rows(self):
        """Sort table rows by country code and zip"""
        commons.LogWrapper.info(
            "Sortiere Daten nach Land und Postleitzahl ..."
        )
        self.dataframe = self.dataframe.sort_values(
            [self.cc_column, self.zip_column]
        )

    def save(self, output_file: paths.Path) -> None:
        """Save the dataframe"""
        del self.dataframe[self.cc_column]
        self.dataframe.to_excel(output_file, index=False)


def process_file(
    source_path: paths.Path,
    target_path: paths.Path,
    user_settings: commons.UserSettings = commons.UserSettings(),
) -> int:
    """Process the file provided in source_path,
    write output to target_path
    and return the appropriate return code
    """
    commons.LogWrapper.info(f"Lade Datei {source_path} …")
    dataframe = pandas.read_excel(source_path)
    commons.LogWrapper.info("… ok")
    data_fixer = DataFixer(dataframe, user_settings)
    statistics = data_fixer.fix_all_zip_codes()
    everything_is_fine, data_changed = commons.evaluate_results(statistics)
    commons.LogWrapper.info(commons.separator_line())
    if data_changed:
        if everything_is_fine:
            try:
                data_fixer.sort_rows()
            except NotImplementedError:
                commons.LogWrapper.info(
                    "Die Daten werden noch nicht nach Postleitzahlen"
                    " sortiert.",
                    "Das muss in Excel/LibreOffice Calc/… durchgeführt"
                    " werden.",
                )
            #
        else:
            commons.LogWrapper.warning(
                "Da die Orignaldaten nicht fehlerfrei waren,"
                " wurden sie nicht nach Land und Postleitzahl sortiert."
            )
        #
        commons.LogWrapper.info(f"Schreibe Ausgabedatei {target_path} …")
        try:
            data_fixer.save(target_path)
        except OSError as error:
            commons.LogWrapper.error(str(error))
            return commons.RETURNCODE_ERROR
        #
        commons.LogWrapper.info("… ok")
    else:
        if everything_is_fine:
            no_errors = "keine Fehler"
        else:
            no_errors = "keine automatisiert behebbaren Fehler"
        #
        commons.LogWrapper.info(
            "Es wird keine Ausgabedatei geschrieben,",
            f"weil die Daten {no_errors} enthalten.",
        )
    #
    return commons.RETURNCODE_OK


# vim: fileencoding=utf-8 sw=4 ts=4 sts=4 expandtab autoindent syntax=python:



#!/usr/bin/env python

"""

postleid.__main__

Locate and fix postal code information in an excel file
Variant using pandas for excel files handling

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import argparse
import gettext
import logging
import shutil
import sys

from typing import List, Optional

# local imports
from postleid import __version__
from postleid import commons
from postleid import fix_excel_files
from postleid import paths
from postleid import presets


def list_supported_countries() -> int:
    """List supported countries and exit with the matching returncode"""
    try:
        country_names = commons.load_country_names_from_file()
    except OSError as error:
        commons.LogWrapper.error(f"{error}")
        return commons.RETURNCODE_ERROR
    #
    try:
        rules = commons.load_rules_from_file()
    except OSError as error:
        commons.LogWrapper.error(f"{error}")
        return commons.RETURNCODE_ERROR
    #
    without_rules: List[str] = []
    commons.LogWrapper.info("Supported countries:", commons.separator_line())
    for iso_cc, names in country_names.items():
        line = f"[{iso_cc}] {' / '.join(names)}"
        if iso_cc in rules:
            print(line)
        else:
            without_rules.append(line)
        #
    #
    if without_rules:
        commons.LogWrapper.debug(
            commons.separator_line(),
            "Countries with names but without rules:",
            commons.separator_line(),
            *without_rules,
        )
    #
    without_cleartext = [
        f"[{iso_cc}] {country_rule.get('comment', '')}"
        for iso_cc, country_rule in rules.items()
        if iso_cc not in country_names
    ]
    if without_cleartext:
        commons.LogWrapper.debug(
            commons.separator_line(),
            "Countries with rules but missing cleartext name(s):",
            commons.separator_line(),
            *without_cleartext,
        )
    #
    return commons.RETURNCODE_OK


def _parse_args(args: Optional[List[str]] = None) -> argparse.Namespace:
    """Parse command line arguments using argparse
    and return the arguments namespace.

    :param args: a list of command line arguments,
        or None to parse sys.argv
    :returns: the parsed command line arguments as returned
        by argparse.ArgumentParser().parse_args()
    """
    # ------------------------------------------------------------------
    # Argparse translation code adapted from
    # <https://github.com/s-ball/i18nparse>
    translation = gettext.translation(
        "argparse",
        localedir=paths.PACKAGE_LOCALE_PATH,
        languages=["de"],
        fallback=True,
    )
    argparse._ = translation.gettext  # type: ignore
    argparse.ngettext = translation.ngettext  # type: ignore
    # ------------------------------------------------------------------
    main_parser = argparse.ArgumentParser(
        prog="postleid",
        description="Postleitzahlen in Excel-Dateien korrigieren",
    )
    main_parser.set_defaults(
        loglevel=logging.INFO,
        settings_file=paths.Path(presets.DEFAULT_USER_SETTINGS_FILE_NAME),
    )
    main_parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Version ausgeben und beenden",
    )
    logging_group = main_parser.add_argument_group(
        "Logging-Optionen",
        "steuern die Meldungsausgaben (Standard-Loglevel: INFO)",
    )
    verbosity = logging_group.add_mutually_exclusive_group()
    verbosity.add_argument(
        "-v",
        "--verbose",
        action="store_const",
        const=logging.DEBUG,
        dest="loglevel",
        help="alle Meldungen ausgeben (Loglevel DEBUG)",
    )
    verbosity.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        const=logging.WARNING,
        dest="loglevel",
        help="nur Warnungen und Fehler ausgeben (Loglevel WARNING)",
    )
    main_parser.add_argument(
        "-l",
        "--list-supported-countries",
        action="store_true",
        help="Unterstützte Länder anzeigen"
        " (der Dateiname muss in diesem Fall zwar auch angegeben werden,"
        " wird jedoch ignoriert)",
    )
    main_parser.add_argument(
        "-o",
        "--output-file",
        metavar="AUSGABEDATEI",
        type=paths.Path,
        help="die Ausgabedatei (Standardwert: Name der Original-Exceldatei"
        f" mit vorangestelltem {presets.DEFAULT_FIXED_FILE_PREFIX!r})",
    )
    main_parser.add_argument(
        "-s",
        "--settings-file",
        metavar="EINSTELLUNGSDATEI",
        type=paths.Path,
        help="die Datei mit Benutzereinstellungen"
        " (Standardwert: %(default)s im aktuellen Verzeichnis)",
    )
    main_parser.add_argument(
        "excel_file",
        metavar="EXCELDATEI",
        type=paths.Path,
        help="die Original-Exceldatei",
    )
    return main_parser.parse_args(args)


def main(args: Optional[List[str]] = None) -> int:
    """Check the zip codes in the input file,
    and write fixed results to the output file if necessary.
    """
    arguments = _parse_args(args)
    commons.LogWrapper(arguments.loglevel)
    if arguments.list_supported_countries:
        return list_supported_countries()
    #
    source_path = arguments.excel_file.resolve()
    target_path = arguments.output_file
    if target_path:
        target_path = target_path.resolve()
    else:
        target_path = (
            source_path.parent
            / f"{presets.DEFAULT_FIXED_FILE_PREFIX}{source_path.name}"
        )
    #
    if not arguments.settings_file.exists():
        default_settings_path = (
            paths.PACKAGE_DATA_PATH / "default_user_settings.yaml"
        )
        commons.LogWrapper.info(
            f"Einstellungsdatei {arguments.settings_file}"
            " noch nicht vorhanden",
            f" → erzeuge eine neue aus {default_settings_path} …",
        )
        shutil.copy2(default_settings_path, arguments.settings_file)
        commons.LogWrapper.info("… ok")
    #
    commons.LogWrapper.info(
        f"Lade Einstellungen aus {arguments.settings_file} …"
    )
    loaded_settings = commons.load_yaml_from_path(arguments.settings_file)
    user_settings = commons.UserSettings(**loaded_settings)
    commons.LogWrapper.info("… ok")
    commons.LogWrapper.info(commons.separator_line())
    return fix_excel_files.process_file(
        source_path,
        target_path,
        user_settings=user_settings,
    )


if __name__ == "__main__":
    sys.exit(main())


# vim: fileencoding=utf-8 sw=4 ts=4 sts=4 expandtab autoindent syntax=python:



#!/usr/bin/env python

"""

postleid.commons

Common functions and classes

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import bisect
import dataclasses
import logging
import textwrap

from collections import Counter

from typing import Any, List, Tuple, Union

import yaml

# local imports
from postleid import paths
from postleid import presets


# STATISTICS KEYWORDS
S_UNCHANGED = "(=) unverändert"
S_FIXED = "(✔) korrigiert"
S_MISSING_RULES = "(✘) Regeln fehlen"
S_WRONG_FORMAT = "(✘) falsches Format"
S_WRONG_DATA_TYPE = "(✘) falscher Datentyp"
S_OUT_OF_RANGE = "(✘) außerhalb des Bereichs"

STATES_OK = (S_UNCHANGED, S_FIXED)
STATES_ERROR = (S_WRONG_DATA_TYPE, S_WRONG_FORMAT, S_OUT_OF_RANGE)

# Return codes
RETURNCODE_OK = 0
RETURNCODE_ERROR = 1


class LogWrapper:

    """wraps logging functionality"""

    textwrapper = textwrap.TextWrapper(presets.LOG_MESSAGE_MAX_WIDTH)

    def __init__(self, loglevel: int) -> None:
        """Initialize logging"""
        logging.basicConfig(
            format=presets.LOG_MESSAGE_FORMAT,
            level=loglevel,
        )

    @classmethod
    def log_formatted(cls, level: int, *messages: str) -> None:
        """Log the already formatted messages with loglevel,
        wrapping them to a line length of
        at most preferences.LOG_MESSAGE_MAX_WIDTH characters
        """
        output_lines = []
        for single_message in messages:
            output_lines.extend(cls.textwrapper.wrap(single_message))
        #
        for line in output_lines:
            logging.log(level, line)
        #

    @classmethod
    def info(cls, *messages: str) -> None:
        """Log with INFO level"""
        cls.log_formatted(logging.INFO, *messages)

    @classmethod
    def debug(cls, *messages: str) -> None:
        """Log with DEBUG level"""
        cls.log_formatted(logging.DEBUG, *messages)

    @classmethod
    def error(cls, *messages: str) -> None:
        """Log with ERROR level"""
        cls.log_formatted(logging.ERROR, *messages)

    @classmethod
    def warning(cls, *messages: str) -> None:
        """Log with WARNING level"""
        cls.log_formatted(logging.WARNING, *messages)


@dataclasses.dataclass
class UserSettings:

    """Container object for keeping user settings"""

    default_country_code: str = presets.DEFAULT_CC
    guess_1000s: bool = presets.GUESS_1000S
    country_headings: Union[
        List[str], Tuple[str, ...]
    ] = presets.COUNTRY_HEADINGS
    postal_code_heading_parts: Union[
        List[str], Tuple[str, ...]
    ] = presets.POSTAL_CODE_HEADING_PARTS


def separator_line(
    element: str = "–", width: int = presets.LOG_MESSAGE_MAX_WIDTH
) -> str:
    """Return a separator line"""
    return element * width


def evaluate_results(statistics: List[str]) -> Tuple[bool, bool]:
    """Evaluate results, show a summary,
    and return two flags: "everything is fine" and "data changed"
    """
    statistics.sort()
    results = Counter(statistics)
    data_changed = bool(results[S_FIXED])
    successful: Counter = Counter()
    errors: Counter = Counter()
    for keyword in sorted(set(statistics)):
        if keyword in (S_FIXED, S_UNCHANGED):
            target = successful
        else:
            target = errors
        #
        start = bisect.bisect_left(statistics, keyword)
        end = bisect.bisect_right(statistics, keyword)
        target.update({keyword: end - start})
        #
    #
    everything_is_fine = not errors.total()
    LogWrapper.log_formatted(
        logging.DEBUG if everything_is_fine else logging.WARNING,
        separator_line(),
    )
    max_kw_width = max(len(keyword) for keyword in results)
    LogWrapper.info(
        f"Ergebnis: {results.total()} Datensätze verarbeitet, davon:",
    )
    for category in (successful, errors):
        for keyword, frequency in category.items():
            LogWrapper.log_formatted(
                logging.INFO if keyword in STATES_OK else logging.WARNING,
                f" {f'{keyword}:':<{max_kw_width + 1}} {frequency}",
            )
        #
    #
    return everything_is_fine, data_changed


def load_yaml_from_path(path: paths.Path) -> Any:
    """Load a YAML file from the provided path
    and return its deserialized contents
    """
    return yaml.safe_load(path.read_text(encoding="utf-8"))


def load_yaml_data_file(file_name: Union[str, paths.Path]) -> Any:
    """Load YAML from the provided file
    located in the data subdirectory
    and return its deserialized contents
    """
    return load_yaml_from_path(paths.PACKAGE_DATA_PATH / file_name)


def load_rules_from_file() -> Any:
    """Load the postal code rules by country code
    from the appropriate file
    """
    return load_yaml_data_file("postal_code_rules_by_cc.yaml")


def load_country_names_from_file() -> Any:
    """Load the country names by country code
    from the appropriate file
    """
    return load_yaml_data_file("country_names_by_cc.yaml")


# vim: fileencoding=utf-8 sw=4 ts=4 sts=4 expandtab autoindent syntax=python:



#!/usr/bin/env python

"""

postleid.presets

Presets for postleid

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


# Prefix for the fixed file name
DEFAULT_FIXED_FILE_PREFIX = "fixed-"

# User settings default file name
DEFAULT_USER_SETTINGS_FILE_NAME = "postleid-settings.yaml"

# Logging options
LOG_MESSAGE_FORMAT = "%(levelname)-8s | %(message)s"
LOG_MESSAGE_MAX_WIDTH = 68

#
# Defaults for the commons.UserSettings class
#

# Default country code
DEFAULT_CC = "de"

# Multiply values by 1000 if lower than 1000?
GUESS_1000S = False

# Headings in lower case indicating a country column
COUNTRY_HEADINGS = ("Land", "Staat", "Country")

# Heading parts in lower case indicating a postal code column
POSTAL_CODE_HEADING_PARTS = ("PLZ", "Postleit", "Zip Code")


# vim: fileencoding=utf-8 sw=4 ts=4 sts=4 expandtab autoindent syntax=python:



#!/usr/bin/env python

"""

postleid.rule_checks

Formal postal code checks using different rules per country

Copyright (C) 2023 Rainer Schwarzbach

This file is part of postleid.

postleid is free software: you can redistribute it and/or modify
it under the terms of the MIT License.

postleid is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the LICENSE file for more details.

"""


import dataclasses
import math
import re

from typing import Any, Dict, Iterator, List, Optional, Union

# local module
from postleid import commons


class ValidatorError(Exception):

    """Base class for errors raised by this module"""

    def __init__(self, *args: object) -> None:
        """Initialize the base class"""
        super().__init__(*args)
        self.additional_information: List[str] = []

    def add_info(self, *more_info: str) -> None:
        """provide additional information"""
        self.additional_information.extend(more_info)


class InvalidFormatError(ValidatorError):

    """Raised if a candidate does not mantch any defined format"""


class MissingRulesError(ValidatorError):

    """Raised if rules for the provided country are not defined"""


class OutOfRangeError(ValidatorError):

    """Raised if a zip is out of range"""


class UnsupportedDataTypeError(ValidatorError):

    """Raised if a candidate is of an unsupported data type"""


def get_pattern_data(rules: Dict[str, Any]) -> Dict[str, Any]:
    """Return only the rule parts relevant for pattern interpretation"""
    return {
        key: value
        for key, value in rules.items()
        if key
        in ("pattern", "min", "max", "only", "except", "remove_leading_zeroes")
    }


@dataclasses.dataclass
class PartialValueRules:

    """Store rules for partial values"""

    only: List[str]
    excluded: List[str]
    minimum: Optional[int] = None
    maximum: Optional[int] = None
    pattern_length: int = 0
    remove_leading_zeroes: Optional[bool] = False

    def validated_numeric_part(self, numeric_candidate: int) -> str:
        """Return the validated numeric part as string
        or raise an OutOfRangeError on violations
        """
        if self.minimum is not None and numeric_candidate < self.minimum:
            raise OutOfRangeError(
                f"{numeric_candidate} is below defined minimum"
                f" of {self.minimum}"
            )
        #
        if self.maximum is not None and numeric_candidate > self.maximum:
            raise OutOfRangeError(
                f"{numeric_candidate} is above defined maximum"
                f" of {self.maximum}"
            )
        #
        validated_part = f"{numeric_candidate:0{self.pattern_length}d}"
        if self.remove_leading_zeroes:
            validated_part = validated_part.lstrip("0")
        #
        return validated_part

    def check_limitations(self, candidate_part: str) -> None:
        """Check candidate_part against only / except
        and raise an InvalidFormatError on rule violations
        """
        if self.only and candidate_part not in self.only:
            raise InvalidFormatError(
                f"{candidate_part} must be one of {self.only!r}"
            )
        #
        if self.excluded and candidate_part in self.excluded:
            raise InvalidFormatError(
                f"{candidate_part} must not be one of {self.excluded!r}"
            )
        #


class PatternComponent:

    """Check one component"""

    translations = {
        "[a]": "[A-Z]?",
        "[a/n]": r"[A-Z\d]?",
        "a": "[A-Z]",
        "n": r"\d",
        "?": r"[A-Z\d]",
    }
    prx_all_numeric_pattern = re.compile(r"\An+\Z")
    prx_all_numeric_literal = re.compile(r"\A\d+\Z")

    def __init__(self, country_code: str, **rules) -> None:
        """Initialize rules for the country"""
        # rules: literal | pattern, min, max, only, except
        only: List[str] = []
        excluded: List[str] = []
        minimum: Optional[int] = None
        maximum: Optional[int] = None
        remove_leading_zeroes = False
        if "literal" in rules:
            literal = str(rules["literal"])
            only.append(literal)
            self.pattern = literal
            self.__regex = "." * len(literal)
            self.all_numeric = self.prx_all_numeric_literal.match(literal)
        else:
            self.pattern = rules["pattern"]
            if self.pattern == "cc":
                self.__regex = re.escape(country_code.upper())
            else:
                self.__regex = self.__translate_pattern(self.pattern)
                only.extend(str(item) for item in rules.get("only", []))
                excluded.extend(str(item) for item in rules.get("except", []))
                minimum = rules.get("min")
                maximum = rules.get("max")
                remove_leading_zeroes = rules.get(
                    "remove_leading_zeroes", False
                )
            #
            self.all_numeric = self.prx_all_numeric_pattern.match(self.pattern)
            if (
                minimum is not None
                and self.all_numeric
                and remove_leading_zeroes
            ):
                self.__regex = (
                    f"{self.translations['n']}"
                    f"{{{int(math.log10(minimum))},{len(self.pattern)}}}"
                )
            else:
                remove_leading_zeroes = False
            #
        #
        self.__rules = PartialValueRules(
            only,
            excluded,
            minimum=minimum,
            maximum=maximum,
            pattern_length=len(self.pattern),
            remove_leading_zeroes=remove_leading_zeroes,
        )

    @property
    def regex(self) -> str:
        """Return the partial regex"""
        return self.__regex

    @classmethod
    def __translate_pattern(cls, pattern: str) -> str:
        """Translate a pattern to a regular expression"""
        re_parts: List[str] = []
        try:
            re_parts.append(cls.translations[pattern])
        except KeyError:
            for character in pattern:
                re_parts.append(cls.translations[character])
            #
        #
        return "".join(re_parts)

    def validated(self, candidate_part: str) -> str:
        """Validate a postal code part and raise the appropriate Exception
        if it does not conform to the rules.
        Return the validated part.
        """
        validated_part = candidate_part
        try:
            if self.all_numeric:
                validated_part = self.__rules.validated_numeric_part(
                    int(candidate_part, 10)
                )
            #
            self.__rules.check_limitations(validated_part)
        except ValidatorError as error:
            error.add_info(f"Pattern component is {self.pattern!r}")
            raise error
        #
        return validated_part


class VariantChecker:

    """Check one variant"""

    def __init__(self, country_code: str, **rules) -> None:
        """Initialize rules for the country"""
        self.__cc = country_code
        # rules: compound | literal | pattern
        components = rules.get("compound", [])
        self.__components: List[PatternComponent] = []
        if components:
            self.__components.extend(
                PatternComponent(self.__cc, **single_component)
                for single_component in components
            )
        else:
            if "literal" in rules:
                self.__components.append(
                    PatternComponent(self.__cc, literal=rules["literal"])
                )
            elif "pattern" in rules:
                pattern_data = get_pattern_data(rules)
                self.__components.append(
                    PatternComponent(self.__cc, **pattern_data)
                )
            #
        #
        # store the full pattern and regex
        self.__full_pattern = "".join(part.pattern for part in self)
        self.__full_prx = re.compile(
            f"\\A({')('.join(part.regex for part in self)})\\Z"
        )
        self.__all_numeric = all(part.all_numeric for part in self)

    @property
    def full_pattern(self) -> str:
        """Return the full pattern"""
        return self.__full_pattern

    def __getitem__(self, index: int) -> PatternComponent:
        """Return pattern component at index"""
        return self.__components[index]

    def __iter__(self) -> Iterator[PatternComponent]:
        """Return an iterator over the pattern components"""
        return iter(self.__components)

    def __len__(self) -> int:
        """Return the number of pattern components"""
        return len(self.__components)

    def format_post_code(
        self,
        candidate: Union[int, str],
    ) -> str:
        """Return a validated and correctly formatted postal code"""
        # split into components and let the component checkers do the work
        if isinstance(candidate, int):
            if self.__all_numeric:
                candidate = f"{candidate:0{len(self.full_pattern)}d}"
            else:
                raise InvalidFormatError("Not all numeric pattern")
            #
        else:
            candidate = candidate.upper()
        #
        matched = self.__full_prx.match(candidate)
        if not matched:
            raise InvalidFormatError(
                f"{candidate!r} does not match pattern {self.full_pattern!r}"
            )
        #
        validated_parts: List[str] = []
        for c_index, group in enumerate(matched.groups()):
            try:
                validated_parts.append(self[c_index].validated(group))
            except ValidatorError as error:
                error.add_info(
                    f"Candidate {candidate!r} examined part was {group!r}",
                    f"Component #{c_index +1} of {len(self)} in variant",
                )
                raise error
        #
        return "".join(validated_parts)


class CodeChecker:

    """Post code formal check for one country"""

    def __init__(self, country_code: str, **rules) -> None:
        """Initialize rules for the country"""
        self.__cc = country_code
        # rules: comment, urls, variants | literal | pattern | compound
        self.__comment = rules.get("comment", f"{country_code!r} code checker")
        self.__urls = rules.get("urls", [])
        variants = rules.get("variants", [])
        self.__variants: List[VariantChecker] = []
        if variants:
            self.__variants.extend(
                VariantChecker(self.__cc, **single_variant)
                for single_variant in variants
            )
        else:
            if "literal" in rules:
                self.__variants.append(
                    VariantChecker(self.__cc, literal=rules["literal"])
                )
            elif "pattern" in rules:
                pattern_data = get_pattern_data(rules)
                self.__variants.append(
                    VariantChecker(self.__cc, **pattern_data)
                )
            elif "compound" in rules:
                self.__variants.append(
                    VariantChecker(self.__cc, compound=rules["compound"])
                )
            #
        #

    def __iter__(self) -> Iterator[VariantChecker]:
        """Return an iterator over the variants"""
        return iter(self.__variants)

    def format_post_code(
        self,
        candidate: Union[int, str],
    ) -> str:
        """Return a validated and correctly formatted postal code"""
        if not isinstance(candidate, (int, str)):
            raise UnsupportedDataTypeError(
                f"Cannot check {candidate!r},"
                " only integers and strings are supported."
            )
        #
        last_error: ValidatorError = MissingRulesError()
        for v_index, single_variant in enumerate(self):
            try:
                return single_variant.format_post_code(candidate)
            except ValidatorError as error:
                last_error = error
                last_error.add_info(
                    f"Variant #{v_index + 1} of {len(self.__variants)}",
                    f"Full pattern is {single_variant.full_pattern!r}",
                )
            #
        #
        last_error.add_info(
            f"Country: {self.__comment}",
            *[f"<{single_url}>" for single_url in self.__urls],
        )
        raise last_error


class ValidatorsCache:

    """Cache for validators

    load rules for all countries, build and cache rules per country
    lazily on demand.
    """

    def __init__(self, default_cc: str = "us") -> None:
        """Initialize the CodeCheckers cache"""
        self.__rules = commons.load_rules_from_file()
        self.__cache: Dict[str, CodeChecker] = {}
        self.__default_cc = default_cc

    def __getitem__(self, name: str) -> CodeChecker:
        """Return cached country code checker"""
        try:
            return self.__cache[name]
        except KeyError:
            try:
                country_rules = self.__rules[name]
            except KeyError as error:
                raise MissingRulesError from error
            #
            return self.__cache.setdefault(
                name,
                CodeChecker(name, **country_rules),
            )
        #

    def output_validated(
        self,
        candidate: Union[float, int, str],
        country: Optional[str] = None,
    ) -> str:
        """Output a validated post code for the country"""
        if country is None:
            country = self.__default_cc
        #
        if isinstance(candidate, float):
            candidate = int(candidate)
        #
        return self[country].format_post_code(candidate)


# vim: fileencoding=utf-8 sw=4 ts=4 sts=4 expandtab autoindent syntax=python:
